import asyncio
import functools
from memorix_client_redis.hash_key import hash_key
from memorix_client_redis.json import from_json, to_json, bytes_to_str
import typing
from memorix_client_redis.memorix_base import MemorixBase
from .cache_options import CacheOptions

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class CacheItem(typing.Generic[KT, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
        options: typing.Optional[CacheOptions] = None,
    ) -> None:
        self._api = api
        self._id = id
        self._payload_class = payload_class
        self._options = options
        self._has_key = True

    def _key(self, key: KT) -> str:
        return hash_key(api=self._api, id=self._id, key=key, has_key=self._has_key)

    def _extend(
        self,
        key: KT,
    ) -> None:
        if (
            self._options is None
            or self._options.ttl is None
            or self._options.ttl == "0"
        ):
            return
        ttl = int(self._options.ttl)

        hashed_key = self._key(key=key)

        self._api._connection.redis.expire(
            hashed_key,
            ttl,
        )

    async def _async_extend(
        self,
        key: KT,
    ) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            functools.partial(
                CacheItem._extend,
                self=self,
                key=key,
            ),
        )

    def _get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        data_bytes = self._api._connection.redis.get(
            self._key(key=key),
        )
        if data_bytes is None:
            return None
        if (
            self._options is not None
            and self._options.extend_on_get is not None
            and self._options.extend_on_get == "true"
        ):
            CacheItem._extend(self=self, key=key)

        data_str = bytes_to_str(data_bytes)
        payload = from_json(value=data_str, data_class=self._payload_class)
        return payload

    async def _async_get(self, key: KT) -> typing.Optional[PT]:
        loop = asyncio.get_running_loop()
        payload = await loop.run_in_executor(
            None,
            functools.partial(
                CacheItem._get,
                self=self,
                key=key,
            ),
        )
        return payload

    def _set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        payload_json = to_json(payload)
        return self._api._connection.redis.set(
            self._key(key=key),
            payload_json,
            ex=int(self._options.ttl)
            if self._options is not None
            and self._options.ttl is not None
            and self._options.ttl != "0"
            else None,
        )

    async def _async_set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        loop = asyncio.get_running_loop()
        res = await loop.run_in_executor(
            None,
            functools.partial(
                CacheItem._set,
                self=self,
                key=key,
                payload=payload,
            ),
        )
        return res

    def _delete(
        self,
        key: KT,
    ) -> None:
        self._api._connection.redis.delete(
            self._key(key=key),
        )

    async def _async_delete(
        self,
        key: KT,
    ) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            functools.partial(
                CacheItem._delete,
                self=self,
                key=key,
            ),
        )


class CacheItemTTT(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return self._get(key=key)

    async def async_get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return await self._async_get(key=key)

    def set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set(key=key, payload=payload)

    async def async_set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set(key=key, payload=payload)

    def delete(
        self,
        key: KT,
    ) -> None:
        self._delete(key=key)

    async def async_delete(
        self,
        key: KT,
    ) -> None:
        await self._async_delete(key=key)


class CacheItemTTF(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return self._get(key=key)

    async def async_get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return await self._async_get(key=key)

    def set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set(key=key, payload=payload)

    async def async_set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set(key=key, payload=payload)


class CacheItemTFT(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return self._get(key=key)

    async def async_get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return await self._async_get(key=key)

    def delete(
        self,
        key: KT,
    ) -> None:
        self._delete(key=key)

    async def async_delete(
        self,
        key: KT,
    ) -> None:
        await self._async_delete(key=key)


class CacheItemTFF(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return self._get(key=key)

    async def async_get(
        self,
        key: KT,
    ) -> typing.Optional[PT]:
        return await self._async_get(key=key)


class CacheItemFTT(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set(key=key, payload=payload)

    async def async_set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set(key=key, payload=payload)

    def delete(
        self,
        key: KT,
    ) -> None:
        self._delete(key=key)

    async def async_delete(
        self,
        key: KT,
    ) -> None:
        await self._async_delete(key=key)


class CacheItemFTF(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set(key=key, payload=payload)

    async def async_set(
        self,
        key: KT,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set(key=key, payload=payload)


class CacheItemFFT(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)

    def delete(
        self,
        key: KT,
    ) -> None:
        self._delete(key=key)

    async def async_delete(
        self,
        key: KT,
    ) -> None:
        await self._async_delete(key=key)


class CacheItemFFF(CacheItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def extend(self, key: KT) -> None:
        self._extend(key)

    async def async_extend(self, key: KT) -> None:
        await self._async_extend(key)


class CacheItemNoKey(CacheItem[None, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
        options: typing.Optional[CacheOptions] = None,
    ) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class, options=options)
        self._has_key = False

    def _key_no_key(self) -> str:
        return CacheItem._key(self, key=None)

    def _extend_no_key(
        self,
    ) -> None:
        CacheItem._extend(self, key=None)

    async def _async_extend_no_key(
        self,
    ) -> None:
        await CacheItem._async_extend(
            self,
            key=None,
        )

    def _get_no_key(self) -> typing.Optional[PT]:
        return CacheItem._get(self, key=None)

    async def _async_get_no_key(self) -> typing.Optional[PT]:
        return await CacheItem._async_get(self, key=None)  # type: ignore

    def _set_no_key(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return CacheItem._set(
            self,
            key=None,
            payload=payload,
        )

    async def _async_set_no_key(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await CacheItem._async_set(
            self,
            key=None,
            payload=payload,
        )

    def _delete_no_key(
        self,
    ) -> None:
        CacheItem._delete(self, key=None)

    async def _async_delete_no_key(
        self,
    ) -> None:
        await CacheItem._async_delete(
            self,
            key=None,
        )


class CacheItemTTTNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def get(
        self,
    ) -> typing.Optional[PT]:
        return self._get_no_key()

    async def async_get(
        self,
    ) -> typing.Optional[PT]:
        return await self._async_get_no_key()

    def set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set_no_key(payload=payload)

    async def async_set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set_no_key(payload=payload)

    def delete(
        self,
    ) -> None:
        self._delete_no_key()

    async def async_delete(
        self,
    ) -> None:
        await self._async_delete_no_key()


class CacheItemTTFNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def get(
        self,
    ) -> typing.Optional[PT]:
        return self._get_no_key()

    async def async_get(
        self,
    ) -> typing.Optional[PT]:
        return await self._async_get_no_key()

    def set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set_no_key(payload=payload)

    async def async_set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set_no_key(payload=payload)


class CacheItemTFTNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def get(
        self,
    ) -> typing.Optional[PT]:
        return self._get_no_key()

    async def async_get(
        self,
    ) -> typing.Optional[PT]:
        return await self._async_get_no_key()

    def delete(
        self,
    ) -> None:
        self._delete_no_key()

    async def async_delete(
        self,
    ) -> None:
        await self._async_delete_no_key()


class CacheItemTFFNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def get(
        self,
    ) -> typing.Optional[PT]:
        return self._get_no_key()

    async def async_get(
        self,
    ) -> typing.Optional[PT]:
        return await self._async_get_no_key()


class CacheItemFTTNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set_no_key(payload=payload)

    async def async_set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set_no_key(payload=payload)

    def delete(
        self,
    ) -> None:
        self._delete_no_key()

    async def async_delete(
        self,
    ) -> None:
        await self._async_delete_no_key()


class CacheItemFTFNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return self._set_no_key(payload=payload)

    async def async_set(
        self,
        payload: PT,
    ) -> typing.Optional[bool]:
        return await self._async_set_no_key(payload=payload)


class CacheItemFFTNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()

    def delete(
        self,
    ) -> None:
        self._delete_no_key()

    async def async_delete(
        self,
    ) -> None:
        await self._async_delete_no_key()


class CacheItemFFFNoKey(CacheItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def extend(self) -> None:
        self._extend_no_key()

    async def async_extend(self) -> None:
        await self._async_extend_no_key()



from .memorix_base import MemorixBase as MemorixBase
from .cache import All as MemorixCacheAll
from .pubsub import All as MemorixPubSubAll
from .task import All as MemorixTaskAll
from dataclasses import dataclass as dataclass



import typing
import typing_extensions
from .task_base import TaskBase
from .task_options import TaskOptions
from .task_item import (
    TaskItemTTTT,
    TaskItemTTTF,
    TaskItemTTFT,
    TaskItemTTFF,
    TaskItemTFTT,
    TaskItemTFTF,
    TaskItemTFFT,
    TaskItemTFFF,
    TaskItemFTTT,
    TaskItemFTTF,
    TaskItemFTFT,
    TaskItemFTFF,
    TaskItemFFTT,
    TaskItemFFTF,
    TaskItemFFFT,
    TaskItemFFFF,
    TaskItemTTTTNoKey,
    TaskItemTTTFNoKey,
    TaskItemTTFTNoKey,
    TaskItemTTFFNoKey,
    TaskItemTFTTNoKey,
    TaskItemTFTFNoKey,
    TaskItemTFFTNoKey,
    TaskItemTFFFNoKey,
    TaskItemFTTTNoKey,
    TaskItemFTTFNoKey,
    TaskItemFTFTNoKey,
    TaskItemFTFFNoKey,
    TaskItemFFTTNoKey,
    TaskItemFFTFNoKey,
    TaskItemFFFTNoKey,
    TaskItemFFFFNoKey,
)

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class All(object):
    Base: typing_extensions.TypeAlias = TaskBase
    Options: typing_extensions.TypeAlias = TaskOptions

    ItemTTTT: typing_extensions.TypeAlias = TaskItemTTTT[KT, PT]
    ItemTTTF: typing_extensions.TypeAlias = TaskItemTTTF[KT, PT]
    ItemTTFT: typing_extensions.TypeAlias = TaskItemTTFT[KT, PT]
    ItemTTFF: typing_extensions.TypeAlias = TaskItemTTFF[KT, PT]
    ItemTFTT: typing_extensions.TypeAlias = TaskItemTFTT[KT, PT]
    ItemTFTF: typing_extensions.TypeAlias = TaskItemTFTF[KT, PT]
    ItemTFFT: typing_extensions.TypeAlias = TaskItemTFFT[KT, PT]
    ItemTFFF: typing_extensions.TypeAlias = TaskItemTFFF[KT, PT]
    ItemFTTT: typing_extensions.TypeAlias = TaskItemFTTT[KT, PT]
    ItemFTTF: typing_extensions.TypeAlias = TaskItemFTTF[KT, PT]
    ItemFTFT: typing_extensions.TypeAlias = TaskItemFTFT[KT, PT]
    ItemFTFF: typing_extensions.TypeAlias = TaskItemFTFF[KT, PT]
    ItemFFTT: typing_extensions.TypeAlias = TaskItemFFTT[KT, PT]
    ItemFFTF: typing_extensions.TypeAlias = TaskItemFFTF[KT, PT]
    ItemFFFT: typing_extensions.TypeAlias = TaskItemFFFT[KT, PT]
    ItemFFFF: typing_extensions.TypeAlias = TaskItemFFFF[KT, PT]
    ItemTTTTNoKey: typing_extensions.TypeAlias = TaskItemTTTTNoKey[PT]
    ItemTTTFNoKey: typing_extensions.TypeAlias = TaskItemTTTFNoKey[PT]
    ItemTTFTNoKey: typing_extensions.TypeAlias = TaskItemTTFTNoKey[PT]
    ItemTTFFNoKey: typing_extensions.TypeAlias = TaskItemTTFFNoKey[PT]
    ItemTFTTNoKey: typing_extensions.TypeAlias = TaskItemTFTTNoKey[PT]
    ItemTFTFNoKey: typing_extensions.TypeAlias = TaskItemTFTFNoKey[PT]
    ItemTFFTNoKey: typing_extensions.TypeAlias = TaskItemTFFTNoKey[PT]
    ItemTFFFNoKey: typing_extensions.TypeAlias = TaskItemTFFFNoKey[PT]
    ItemFTTTNoKey: typing_extensions.TypeAlias = TaskItemFTTTNoKey[PT]
    ItemFTTFNoKey: typing_extensions.TypeAlias = TaskItemFTTFNoKey[PT]
    ItemFTFTNoKey: typing_extensions.TypeAlias = TaskItemFTFTNoKey[PT]
    ItemFTFFNoKey: typing_extensions.TypeAlias = TaskItemFTFFNoKey[PT]
    ItemFFTTNoKey: typing_extensions.TypeAlias = TaskItemFFTTNoKey[PT]
    ItemFFTFNoKey: typing_extensions.TypeAlias = TaskItemFFTFNoKey[PT]
    ItemFFFTNoKey: typing_extensions.TypeAlias = TaskItemFFFTNoKey[PT]
    ItemFFFFNoKey: typing_extensions.TypeAlias = TaskItemFFFFNoKey[PT]



import typing
import typing_extensions
from .cache_base import CacheBase
from .cache_options import CacheOptions
from .cache_item import (
    CacheItemTTT,
    CacheItemTTF,
    CacheItemTFT,
    CacheItemTFF,
    CacheItemFTT,
    CacheItemFTF,
    CacheItemFFT,
    CacheItemFFF,
    CacheItemTTTNoKey,
    CacheItemTTFNoKey,
    CacheItemTFTNoKey,
    CacheItemTFFNoKey,
    CacheItemFTTNoKey,
    CacheItemFTFNoKey,
    CacheItemFFTNoKey,
    CacheItemFFFNoKey,
)

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class All(object):
    Base: typing_extensions.TypeAlias = CacheBase
    Options: typing_extensions.TypeAlias = CacheOptions
    ItemTTT: typing_extensions.TypeAlias = CacheItemTTT[KT, PT]
    ItemTTF: typing_extensions.TypeAlias = CacheItemTTF[KT, PT]
    ItemTFT: typing_extensions.TypeAlias = CacheItemTFT[KT, PT]
    ItemTFF: typing_extensions.TypeAlias = CacheItemTFF[KT, PT]
    ItemFTT: typing_extensions.TypeAlias = CacheItemFTT[KT, PT]
    ItemFTF: typing_extensions.TypeAlias = CacheItemFTF[KT, PT]
    ItemFFT: typing_extensions.TypeAlias = CacheItemFFT[KT, PT]
    ItemFFF: typing_extensions.TypeAlias = CacheItemFFF[KT, PT]
    ItemTTTNoKey: typing_extensions.TypeAlias = CacheItemTTTNoKey[PT]
    ItemTTFNoKey: typing_extensions.TypeAlias = CacheItemTTFNoKey[PT]
    ItemTFTNoKey: typing_extensions.TypeAlias = CacheItemTFTNoKey[PT]
    ItemTFFNoKey: typing_extensions.TypeAlias = CacheItemTFFNoKey[PT]
    ItemFTTNoKey: typing_extensions.TypeAlias = CacheItemFTTNoKey[PT]
    ItemFTFNoKey: typing_extensions.TypeAlias = CacheItemFTFNoKey[PT]
    ItemFFTNoKey: typing_extensions.TypeAlias = CacheItemFFTNoKey[PT]
    ItemFFFNoKey: typing_extensions.TypeAlias = CacheItemFFFNoKey[PT]



import typing


class CacheOptions(object):
    def __init__(
        self,
        ttl: typing.Optional[str] = None,
        extend_on_get: typing.Optional[str] = None,
    ) -> None:
        self.ttl = ttl
        self.extend_on_get = extend_on_get



import asyncio
import functools
from memorix_client_redis.hash_key import hash_key
from memorix_client_redis.json import from_json, to_json, bytes_to_str
import typing
from memorix_client_redis.memorix_base import MemorixBase
from .task_options import TaskOptions

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class TaskItemQueue(object):
    def __init__(self, queue_size: int) -> None:
        self.queue_size = queue_size


class TaskItem(typing.Generic[KT, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
        options: typing.Optional[TaskOptions] = None,
    ) -> None:
        self._api = api
        self._id = id
        self._payload_class = payload_class
        self._options = options
        self._has_key = True

    def _key(self, key: KT) -> str:
        return hash_key(api=self._api, id=self._id, key=key, has_key=self._has_key)

    def _enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        wrapped_payload_json = to_json(payload)

        queue_size = self._api._connection.redis.rpush(
            self._key(key=key),
            wrapped_payload_json,
        )
        return TaskItemQueue(queue_size=queue_size)

    async def _async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            functools.partial(
                TaskItem._enqueue,
                self=self,
                key=key,
            ),
        )

    def _dequeue(
        self,
        key: KT,
    ) -> typing.Generator[PT, None, None]:

        while True:
            if self._options is not None and self._options.queue_type == "lifo":
                [channel_bytes, data_bytes] = self._api._connection.redis.brpop(
                    self._key(key=key),
                )
            else:
                [channel_bytes, data_bytes] = self._api._connection.redis.blpop(
                    self._key(key=key),
                )

            payload_str = bytes_to_str(data_bytes)
            payload = from_json(payload_str, self._payload_class)

            yield payload

    async def _async_dequeue(
        self,
        key: KT,
    ) -> typing.AsyncGenerator[PT, None]:
        print("dequeue async")
        yield typing.cast(PT, None)

    def _empty(self, key: KT) -> None:
        self._api._connection.redis.delete(
            self._key(key=key),
        )

    async def _async_empty(self, key: KT) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            None,
            functools.partial(
                TaskItem._empty,
                self=self,
                key=key,
            ),
        )

    def _get_len(self, key: KT) -> int:
        return self._api._connection.redis.llen(
            self._key(key=key),
        )

    async def _async_get_len(self, key: KT) -> int:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            functools.partial(
                TaskItem._get_len,
                self=self,
                key=key,
            ),
        )


class TaskItemTTTT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemTTTF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)


class TaskItemTTFT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemTTFF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)


class TaskItemTFTT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemTFTF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)


class TaskItemTFFT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemTFFF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return self._enqueue(key=key, payload=payload)

    async def async_enqueue(self, key: KT, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue(key=key, payload=payload)


class TaskItemFTTT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemFTTF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)


class TaskItemFTFT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemFTFF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def dequeue(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._dequeue(key=key)

    async def async_dequeue(self, key: KT) -> typing.AsyncGenerator[PT, None]:
        return self._async_dequeue(key=key)


class TaskItemFFTT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemFFTF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def empty(self, key: KT) -> None:
        self._empty(key=key)

    async def async_empty(self, key: KT) -> None:
        await self._async_empty(key=key)


class TaskItemFFFT(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)

    def get_len(self, key: KT) -> int:
        return self._get_len(key=key)

    async def async_get_len(self, key: KT) -> int:
        return await self._async_get_len(key=key)


class TaskItemFFFF(TaskItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key=key)


class TaskItemNoKey(TaskItem[None, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
        options: typing.Optional[TaskOptions] = None,
    ) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class, options=options)
        self._has_key = False

    def _key_no_key(self) -> str:
        return TaskItem._key(self, key=None)

    def _enqueue_no_key(self, payload: PT) -> TaskItemQueue:
        return TaskItem._enqueue(self, key=None, payload=payload)

    async def _async_enqueue_no_key(self, payload: PT) -> TaskItemQueue:
        return TaskItem._enqueue(self, key=None, payload=payload)

    def _dequeue_no_key(
        self,
    ) -> typing.Generator[PT, None, None]:
        return TaskItem._dequeue(self, key=None)

    async def _async_dequeue_no_key(
        self,
    ) -> typing.AsyncGenerator[PT, None]:
        return TaskItem._async_dequeue(self, key=None)

    def _empty_no_key(self) -> None:
        TaskItem._empty(self, key=None)

    async def _async_empty_no_key(self) -> None:
        await TaskItem._async_empty(self, key=None)

    def _get_len_no_key(self) -> int:
        return TaskItem._get_len(self, key=None)

    async def _async_get_len_no_key(self) -> int:
        return await TaskItem._async_get_len(self, key=None)


class TaskItemTTTTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemTTTFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()


class TaskItemTTFTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemTTFFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()


class TaskItemTFTTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemTFTFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()


class TaskItemTFFTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemTFFFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def enqueue(self, payload: PT) -> TaskItemQueue:
        return self._enqueue_no_key(payload=payload)

    async def async_enqueue(self, payload: PT) -> TaskItemQueue:
        return await self._async_enqueue_no_key(payload=payload)


class TaskItemFTTTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemFTTFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()


class TaskItemFTFTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemFTFFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def dequeue(self) -> typing.Generator[PT, None, None]:
        return self._dequeue_no_key()

    async def async_dequeue(self) -> typing.AsyncGenerator[PT, None]:
        return await self._async_dequeue_no_key()


class TaskItemFFTTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemFFTFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def empty(self) -> None:
        self._empty_no_key()

    async def async_empty(self) -> None:
        await self._async_empty_no_key()


class TaskItemFFFTNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()

    def get_len(self) -> int:
        return self._get_len_no_key()

    async def async_get_len(self) -> int:
        return await self._async_get_len_no_key()


class TaskItemFFFFNoKey(TaskItemNoKey[PT]):
    def key(self) -> str:
        return self._key_no_key()



import asyncio
import functools
from memorix_client_redis.hash_key import hash_key
from memorix_client_redis.json import from_json, to_json, bytes_to_str
from memorix_client_redis.memorix_base import MemorixBase
import typing

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class PubSubItemPublish(object):
    def __init__(
        self,
        subscribers_size: int,
    ) -> None:
        self.subscribers_size = subscribers_size


class PubSubItem(typing.Generic[KT, PT]):
    def __init__(
        self,
        api: MemorixBase,
        id: str,
        payload_class: typing.Type[PT],
    ) -> None:
        self._api = api
        self._id = id
        self._payload_class = payload_class
        self._has_key = True

    def _key(self, key: KT) -> str:
        return hash_key(api=self._api, id=self._id, has_key=self._has_key, key=key)

    def _publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        payload_json = to_json(payload)
        subscribers_size = self._api._connection.redis.publish(
            self._key(key=key),
            payload_json,
        )
        return PubSubItemPublish(subscribers_size=subscribers_size)

    async def _async_publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        loop = asyncio.get_running_loop()
        res = await loop.run_in_executor(
            None,
            functools.partial(
                PubSubItem._publish,
                self=self,
                key=key,
                payload=payload,
            ),
        )
        return res

    def _subscribe(self, key: KT) -> typing.Generator[PT, None, None]:
        sub = self._api._connection.redis.pubsub()
        sub.subscribe(self._key(key=key))
        for message in typing.cast(
            typing.Generator[typing.Dict[str, typing.Union[int, bytes]], None, None],
            sub.listen(),  # type: ignore
        ):
            data_bytes = message["data"]
            if isinstance(data_bytes, bytes):
                data_str = bytes_to_str(data_bytes)
                data = from_json(value=data_str, data_class=self._payload_class)
                yield data

    async def _async_subscribe(
        self,
        key: KT,
    ) -> typing.AsyncGenerator[PT, None]:
        sub = self._api._connection.redis.pubsub()
        sub.subscribe(self._key(key=key))
        loop = asyncio.get_running_loop()
        while True:
            message = await loop.run_in_executor(
                None,
                functools.partial(
                    sub.get_message,
                    ignore_subscribe_messages=True,
                ),
            )
            if message is not None:
                data_bytes = message["data"]
                if isinstance(data_bytes, bytes):
                    data_str = bytes_to_str(data_bytes)
                    data = from_json(value=data_str, data_class=self._payload_class)
                    yield data


class PubSubItemTT(PubSubItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        return self._publish(key=key, payload=payload)

    async def async_publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        return await self._async_publish(key=key, payload=payload)

    def subscribe(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._subscribe(key=key)

    async def async_subscribe(
        self,
        key: KT,
    ) -> typing.AsyncGenerator[PT, None]:
        return self._async_subscribe(key=key)


class PubSubItemTF(PubSubItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        return self._publish(key=key, payload=payload)

    async def async_publish(self, key: KT, payload: PT) -> PubSubItemPublish:
        return await self._async_publish(key=key, payload=payload)


class PubSubItemFT(PubSubItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)

    def subscribe(self, key: KT) -> typing.Generator[PT, None, None]:
        return self._subscribe(key=key)

    async def async_subscribe(
        self,
        key: KT,
    ) -> typing.AsyncGenerator[PT, None]:
        return self._async_subscribe(key=key)


class PubSubItemFF(PubSubItem[KT, PT]):
    def key(self, key: KT) -> str:
        return self._key(key)


class PubSubItemTTNoKey(PubSubItem[None, PT]):
    def __init__(self, api: MemorixBase, id: str, payload_class: type[PT]) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class)
        self._has_key = False

    def key(self) -> str:
        return PubSubItem._key(self, key=None)

    def publish(self, payload: PT) -> PubSubItemPublish:
        return PubSubItem._publish(self, key=None, payload=payload)

    async def async_publish(self, payload: PT) -> PubSubItemPublish:
        return await PubSubItem._async_publish(self, key=None, payload=payload)

    def subscribe(self) -> typing.Generator[PT, None, None]:
        return PubSubItem._subscribe(self, key=None)

    def async_subscribe(self) -> typing.AsyncGenerator[PT, None]:
        return PubSubItem._async_subscribe(self, key=None)


class PubSubItemTFNoKey(PubSubItem[None, PT]):
    def __init__(self, api: MemorixBase, id: str, payload_class: type[PT]) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class)
        self._has_key = False

    def key(self) -> str:
        return PubSubItem._key(self, key=None)

    def publish(self, payload: PT) -> PubSubItemPublish:
        return PubSubItem._publish(self, key=None, payload=payload)

    async def async_publish(self, payload: PT) -> PubSubItemPublish:
        return await PubSubItem._async_publish(self, key=None, payload=payload)


class PubSubItemFTNoKey(PubSubItem[None, PT]):
    def __init__(self, api: MemorixBase, id: str, payload_class: type[PT]) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class)
        self._has_key = False

    def key(self) -> str:
        return PubSubItem._key(self, key=None)

    def subscribe(self) -> typing.Generator[PT, None, None]:
        return PubSubItem._subscribe(self, key=None)

    def async_subscribe(self) -> typing.AsyncGenerator[PT, None]:
        return PubSubItem._async_subscribe(self, key=None)


class PubSubItemFFNoKey(PubSubItem[None, PT]):
    def __init__(self, api: MemorixBase, id: str, payload_class: type[PT]) -> None:
        super().__init__(api=api, id=id, payload_class=payload_class)
        self._has_key = False

    def key(self) -> str:
        return PubSubItem._key(self, key=None)



# flake8: noqa
import typing
import os

if typing.TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from memorix_client_redis import dataclass

from enum import Enum
from memorix_client_redis import (
    MemorixBase,
    MemorixCacheAll,
    MemorixPubSubAll,
    MemorixTaskAll,
)


class Animal(str, Enum):
    dog = "dog"
    cat = "cat"
    person = "person"


@dataclass
class InlineTypeUser(object):
    name: str
    age: typing.Optional[int]


User = InlineTypeUser


class MemorixCache(MemorixCacheAll.Base):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.favoriteAnimal = MemorixCacheAll.ItemTTT[str, Animal](
            api=api,
            id="favoriteAnimal",
            payload_class=Animal,
        )
        self.user = MemorixCacheAll.ItemTTT[str, User](
            api=api,
            id="user",
            payload_class=User,
        )


class MemorixPubSub(MemorixPubSubAll.Base):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.message = MemorixPubSubAll.ItemTTNoKey[str](
            api=api,
            id="message",
            payload_class=str,
        )


class MemorixTask(MemorixTaskAll.Base):
    def __init__(self, api: MemorixBase) -> None:
        super().__init__(api=api)

        self.runAlgo = MemorixTaskAll.ItemTTTTNoKey[str](
            api=api,
            id="runAlgo",
            payload_class=str,
        )


class Memorix(MemorixBase):
    def __init__(self) -> None:
        super().__init__(redis_url=os.environ["REDIS_URL"])

        self._namespace_name_tree = []
        self.cache = MemorixCache(self)
        self.pubsub = MemorixPubSub(self)
        self.task = MemorixTask(self)



import typing
import typing_extensions
from .pubsub_base import PubSubBase
from .pubsub_item import (
    PubSubItemTT,
    PubSubItemTF,
    PubSubItemFT,
    PubSubItemFF,
    PubSubItemTTNoKey,
    PubSubItemTFNoKey,
    PubSubItemFTNoKey,
    PubSubItemFFNoKey,
)

KT = typing.TypeVar("KT")
PT = typing.TypeVar("PT")


class All(object):
    Base: typing_extensions.TypeAlias = PubSubBase
    ItemTT: typing_extensions.TypeAlias = PubSubItemTT[KT, PT]
    ItemTF: typing_extensions.TypeAlias = PubSubItemTF[KT, PT]
    ItemFT: typing_extensions.TypeAlias = PubSubItemFT[KT, PT]
    ItemFF: typing_extensions.TypeAlias = PubSubItemFF[KT, PT]
    ItemTTNoKey: typing_extensions.TypeAlias = PubSubItemTTNoKey[PT]
    ItemTFNoKey: typing_extensions.TypeAlias = PubSubItemTFNoKey[PT]
    ItemFTNoKey: typing_extensions.TypeAlias = PubSubItemFTNoKey[PT]
    ItemFFNoKey: typing_extensions.TypeAlias = PubSubItemFFNoKey[PT]



import typing
from memorix_client_redis.json import to_json
from memorix_client_redis.memorix_base import MemorixBase


def hash_key(api: MemorixBase, id: str, has_key: bool, key: typing.Any) -> str:
    if has_key is False:
        return to_json(value=[*api._namespace_name_tree, id])
    return to_json(
        value=[*api._namespace_name_tree, id, key],
        sort_dict=True,
    )


def hash_key_better(api: MemorixBase, id: str, has_key: bool, key: typing.Any) -> str:
    if has_key:
        return to_json(
            value=[*api._namespace_name_tree, id, key],
            sort_dict=True,
        )
    return to_json(value=[*api._namespace_name_tree, id])
